import 'package:flutter/material.dart';
import 'package:nav_router/nav_router.dart';
import 'package:tivio_sdk/tivio.dart';
import 'package:flutter/widgets.dart';
import 'package:sentry_flutter/sentry_flutter.dart';
import 'package:flutter_dotenv/flutter_dotenv.dart';
import 'pages/player_page.dart';
import 'pages/traverse_data/tivio_widgets_page.dart';
import 'pages/get_data_by_id_page.dart';
import 'pages/recent_videos_page.dart';

Future<void> main() async {
  await DotEnv().load('local.env');

  if (DotEnv().env['SENTRY_ENABLED'] == 'true') {
    await SentryFlutter.init(
      (options) {
        options.dsn = DotEnv().env['SENTRY_DNS'];
      },
      appRunner: () => runApp(ExampleApp()),
    );
  } else {
    runApp(ExampleApp());
  }
}

class ExampleApp extends StatefulWidget {
  @override
  _ExampleAppState createState() => _ExampleAppState();
}

class _ExampleAppState extends State<ExampleApp> {
  bool _isTivioInitialized = false;
  Tivio _tivio;

  @override
  void initState() {
    super.initState();
    initializeTivio();
  }

  initializeTivio() async {
    if (DotEnv().env['TIVIO_SECRET'] == null || DotEnv().env['TIVIO_SECRET'].isEmpty) {
      throw new Exception('Copy .env to local.env and add your TIVIO_SECRET to it');
    }

    _tivio = await initTivio(
      secret: DotEnv().env['TIVIO_SECRET'],
      language: TivioLanguages.cs,
      development: (DotEnv().env['INTERNAL_DEVELOPMENT'] == 'true'),
    );

    /// Widget

    print('subscribeToWidget');
    var widgetStream = await _tivio.subscribeToWidget(widgetId: 'FlutterDemo');

    widgetStream.stream.listen((event) {
      print('subscribeToWidget 1 name: ${event.name}');
    });

    await Future.delayed(Duration(seconds: 1));
    widgetStream.close();
    print('close subscribeToWidget');

    print('subscribeToWidget 2');
    var widgetStream2 = await _tivio.subscribeToWidget(widgetId: 'FlutterDemo');

    widgetStream2.stream.listen((event) {
      print('subscribeToWidget 2 name: ${event.name}');
    });

    await Future.delayed(Duration(seconds: 1));
    widgetStream2.close();
    print('close subscribeToWidget 2');

    /// Channel

    print('subscribeToChannel');
    var channelStream = await _tivio.subscribeToChannel(channelId: 'LzqDUHWGqEPUn6LcEjYN');
    channelStream.stream.listen((event) {
      print('subscribeToChannel name: ${event.name}');
    });

    await Future.delayed(Duration(seconds: 1));
    channelStream.close();
    print('close subscribeToChannel');

    print('subscribeToChannelsInWidget');
    var channelsStream = await _tivio.subscribeToChannelsInWidget(widgetId: 'FlutterDemo');
    channelsStream.stream.listen((event) {
      event.records.forEach((element) {
        print('subscribeToChannelsInWidget name: ${element.name}');
      });
    });

    await Future.delayed(Duration(seconds: 1));
    channelsStream.close();
    print('close subscribeToChannelsInWidget');

    print('subscribeToChannelsInWidget 2');
    var channelsStream2 = await _tivio.subscribeToChannelsInWidget(widgetId: 'FlutterDemo');
    channelsStream2.stream.listen((event) {
      event.records.forEach((element) {
        print('subscribeToChannelsInWidget 2 name: ${element.name}');
      });
    });

    await Future.delayed(Duration(seconds: 1));
    channelsStream2.close();
    print('close subscribeToChannelsInWidget 2');

    /// Section

    print('subscribeToSection');
    var sectionStream = await _tivio.subscribeToSection(sectionId: 'Lhkb2QD0Sqg40InUPzfy');
    sectionStream.stream.listen((event) {
      print('subscribeToSection name: ${event.name}');
    });

    await Future.delayed(Duration(seconds: 1));
    sectionStream.close();
    print('close subscribeToSection');

    print('subscribeToSectionsInChannel');
    var sectionsStream = await _tivio.subscribeToSectionsInChannel(channelId: 'LzqDUHWGqEPUn6LcEjYN');
    sectionsStream.stream.listen((event) {
      event.records.forEach((element) {
        print('subscribeToSectionsInChannel name: ${element.name}');
      });
    });

    await Future.delayed(Duration(seconds: 1));
    sectionsStream.close();
    print('close subscribeToSectionsInChannel');

    /// Video

    print('subscribeToVideo');
    var videoStream = await _tivio.subscribeToVideo(videoId: 'NZAPzzAkg45sUxd2bwuK');
    videoStream.stream.listen((event) {
      print('subscribeToVideo name: ${event.name}');
    });

    await Future.delayed(Duration(seconds: 1));
    videoStream.close();
    print('close subscribeToVideo');

    print('subscribeToVideosInSection');
    var videosStream = await _tivio.subscribeToVideosInSection(sectionId: 'Lhkb2QD0Sqg40InUPzfy');
    videosStream.stream.listen((event) {
      print('subscribeToVideosInSection records length: ${event.records.length}');
      event.records.forEach((element) {
        print('subscribeToVideosInSection name: ${element.name}');
      });
    });

    await Future.delayed(Duration(seconds: 1));
    print('subscribeToVideosInSection next page');
    await _tivio.fetchNextVideosInSection(sectionId: 'Lhkb2QD0Sqg40InUPzfy');

    await Future.delayed(Duration(seconds: 1));
    videosStream.close();
    print('close subscribeToVideosInSection');

    /*setState(() {
      _isTivioInitialized = true;
    });*/
  }

  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'Tivio SDK examples',
      theme: ThemeData(
        primarySwatch: Colors.blue,
        visualDensity: VisualDensity.adaptivePlatformDensity,
      ),
      home: Scaffold(
        appBar: AppBar(
          title: Text('Tivio SDK examples'),
        ),
        body: (_isTivioInitialized)
            ? Flex(
                direction: Axis.vertical,
                crossAxisAlignment: CrossAxisAlignment.start,
                children: [
                  TextButton(
                    onPressed: () => _navigateToPage(PlayerPage(_tivio)),
                    child: Text('Player'),
                  ),
                  TextButton(
                    onPressed: () => _navigateToPage(TivioWidgetsPage(_tivio)),
                    child: Text('Traverse data'),
                  ),
                  TextButton(
                    onPressed: () => _navigateToPage(GetDataByIdPage(_tivio)),
                    child: Text('Get data by id'),
                  ),
                  TextButton(
                    onPressed: () => _navigateToPage(RecentVideosPage(_tivio)),
                    child: Text('Recent videos'),
                  ),
                ],
              )
            : Text('initializing tivio'),
      ),
      navigatorKey: navGK,
    );
  }

  _navigateToPage(Widget page) {
    return routePush(page, RouterType.material);
  }
}
