import 'dart:async';
import 'package:meta/meta.dart';
import 'package:flutter/services.dart';
import 'package:interactive_webview/interactive_webview.dart';
import 'package:uuid/uuid.dart';
import 'package:sentry/sentry.dart';
import 'message_broker.dart';
import 'dart:convert';
import 'dto.dart';
import 'player/tivio_player_wrapper.dart';
import './dto_parsers/tivio_data_widget_parser.dart';
import './dto_parsers/tivio_data_channel_parser.dart';
import './dto_parsers/tivio_data_section_parser.dart';
import './dto_parsers/tivio_data_video_parser.dart';

/// @see https://dart.academy/streams-and-sinks-in-dart-and-flutter/
/// @see https://egghead.io/lessons/flutter-capture-and-handle-data-sequences-with-streams-in-dart
class TivioDataStream<T> {
  StreamController<T> _controller;

  TivioDataStream(this._controller);

  Stream<T> get stream {
    return _controller.stream;
  }

  set controller(StreamController<T> controller) {
    _controller = controller;
  }

  close () {
    return _controller.close();
  }
}

class TivioDataStreamController<T> {
  StreamController<T> _controller;
  TivioDataStream<T> tivioDataStream;

  TivioDataStreamController() {
    _controller = StreamController<T>();
    tivioDataStream = TivioDataStream<T>(_controller);
  }

  reopenIfIsClosed() {
    if (_controller.isClosed) {
      _controller = StreamController<T>();
      tivioDataStream.controller = _controller;
    }
  }

  addSafely (T payload) {
    if (!_controller.isClosed) {
      _controller.add(payload);
    }
  }

  close () async {
    await _controller.close();
  }
}

class Bridge {
  static final Bridge _instance = Bridge._internal();
  final _uuid = Uuid();
  final _webView = InteractiveWebView();
  final _jsonEncoder = JsonEncoder();
  final _dataMessageBroker = DataMessageBroker();
  final _callbackMessageBroker = CallbackMessageBroker();
  Completer _bridgeReady;
  var _tivioChannelDataStreamController = TivioDataStreamController<TivioDataPagination<TivioDataChannel>>();
  var _tivioSectionDataStreamController = TivioDataStreamController<TivioDataPagination<TivioDataSection>>();
  var _tivioVideoDataStreamController = TivioDataStreamController<TivioDataPagination<TivioDataVideo>>();

  factory Bridge() {
    return _instance;
  }

  Bridge._internal();

  loadBridgeJs() {
    PlatformAssetBundle loader = PlatformAssetBundle();
    return loader.loadString('packages/tivio_sdk/assets/bridge.build.js');
  }

  Future<void> init(String appId, String language, bool development) async {
    if (_bridgeReady != null) {
      return _bridgeReady.future;
    }

    try {
      _bridgeReady = Completer();
      String bundle = await loadBridgeJs();

      _webView.didReceiveMessage.listen((event) {
        switch (event.data['type']) {
          case 'data':
            _dataMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
          case 'callback':
            _callbackMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
        }
      });

      _webView.stateChanged.listen((event) {
        if (event.type == WebViewState.didFinish) {
          callFunction(
            'init',
            [appId, language, development],
            (payload) => _bridgeReady.complete(),
          );
        }
      });

      _webView.loadHTML('''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <script>
      $bundle
    </script>
  </head>
  <body>
  </body>
</html>
''', baseUrl: 'https://tiv.io');

      return _bridgeReady.future;
    } catch (exception, stackTrace) {
      await Sentry.captureException(
        exception,
        stackTrace: stackTrace,
      );
      throw exception;
    }
  }

  setLanguage(String language) {
    return callFunction('setLanguage', [language], (payload) => null);
  }

  getPlayer(String id) {
    return callFunction<TivioPlayerWrapper>(
      'getPlayer',
      [id],
      (payload) => TivioPlayerWrapper(id, this),
    );
  }

  playerRegister(String id) {
    return callFunction('playerRegister', [id], (payload) => null);
  }

  playerOnSourceChanged(String id, data) {
    return callFunction(
      'playerOnSourceChanged',
      [id, _jsonEncoder.convert(data)],
      (payload) => null,
    );
  }

  playerPlay(String id) {
    return callFunction('playerPlay', [id], (payload) => {});
  }

  playerPause(String id) {
    return callFunction('playerPause', [id], (payload) => {});
  }

  playerSeekTo(String id, num timestampMs) {
    return callFunction('playerSeekTo', [id, timestampMs], (payload) => {});
  }

  playerOnTimeChanged(String id, num timestampMs) {
    return callFunction(
      'playerOnTimeChanged',
      [id, timestampMs],
      (payload) => null,
    );
  }

  playerOnStateChanged(String id, String state) {
    return callFunction('playerOnStateChanged', [id, state], (payload) => null);
  }

  getWidgetById(String widgetId) {
    return callFunction<TivioDataWidget>(
      'getWidgetById',
      [widgetId],
      (payload) => new TivioDataWidget(
        payload['widgetId'],
        payload['name'],
        this,
      ),
    );
  }

  getChannelById(String channelId) {
    return callFunction<TivioDataChannel>(
      'getChannelById',
      [channelId],
      (payload) => TivioDataChannel(
        payload['id'],
        payload['name'],
        payload['headerLogo'],
        payload['header'],
        this,
      ),
    );
  }

  getSectionById(String sectionId) {
    return callFunction<TivioDataSection>(
      'getSectionById',
      [sectionId],
      (payload) => TivioDataSection(
        payload['id'],
        payload['name'],
        this,
      ),
    );
  }

  getVideoById(String videoId) {
    return callFunction<TivioDataVideo>(
      'getVideoById',
      [videoId],
      (record) => TivioDataVideo.fromDynamic(record),
    );
  }

  widgetGetRecentVideos(String widgetId) {
    return callFunction<List<TivioDataVideo>>(
      'widgetGetRecentVideos',
      [widgetId],
      (payload) => TivioDataVideo.fromDynamicList(payload),
    );
  }

  channelGetRecentVideos(String channelId) {
    return callFunction<List<TivioDataVideo>>(
      'channelGetRecentVideos',
      [channelId],
      (payload) => TivioDataVideo.fromDynamicList(payload),
    );
  }

  searchChannels(String query, int limit) {
    return getPaginatedData<TivioDataChannel>(
      'searchChannels',
      [query, limit],
      (record) => new TivioDataChannel(
        record['id'],
        record['name'],
        record['headerLogo'],
        record['header'],
        this,
      ),
    );
  }

  searchVideos(String query, int limit) {
    return getPaginatedData<TivioDataVideo>(
      'searchVideos',
      [query, limit],
      (record) => TivioDataVideo.fromDynamic(record),
    );
  }

  getPaginatedData<T>(
    String functionName,
    List<dynamic> functionArgs,
    T Function(dynamic record) createDto,
  ) {
    return callFunction<TivioDataPagination<T>>(
      functionName,
      functionArgs,
      (payload) {
        return TivioDataPagination<T>(
          payload['records'].map<T>((record) => createDto(record)).toList(),
          payload['hasNextPage'] ?? false,
        );
      },
    );
  }

  Future<T> callFunction<T>(String functionName, List<dynamic> functionArgs, T Function(dynamic payload) onComplete) {
    Completer done = new Completer<T>();

    String messageId = _uuid.v4();
    functionArgs.add(messageId);
    String serializedArgs = _jsonEncoder.convert(functionArgs);

    _webView.evalJavascript('''window.TivioBridge.$functionName($serializedArgs);''');

    _dataMessageBroker.listen(messageId, (payload) => done.complete(onComplete(payload)));

    return done.future;
  }

  onPlayerPlay(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-play', callback);
  }

  onPlayerPause(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-pause', callback);
  }

  onPlayerChangeSource(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-set-source', callback);
  }

  onPlayerSeekTo(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-seek-to', callback);
  }

  subscribeToWidgets(String widgetId, Function(TivioDataWidget) callback) {
    _callbackMessageBroker.listen('watch-widgets', (payload) {
      if (payload['widgetId'] != null && payload['name'] != null) {
        callback(TivioDataWidget(payload['widgetId'], payload['name'], this));
      }
    });

    return callFunction('watchWidgets', [widgetId], (payload) => null);
  }

  unsubscribeToWidgets() {
    _callbackMessageBroker.removeListener('watch-widgets');
  }

  subscribeToChannels(String widgetId, Function(TivioDataChannel) callback) {
    _callbackMessageBroker.listen('watch-channels', (payload) {
      if (payload['name'] != null) {
        callback(TivioDataChannel(payload['id'], payload['name'], payload['headerLogo'], payload['header'], this));
      }
    });

    return callFunction('watchChannels', [widgetId], (payload) => null);
  }

  unsubscribeToChannels() {
    _callbackMessageBroker.removeListener('watch-channels');
  }

  subscribeToSections(String channelId, Function(TivioDataSection) callback) {
    _callbackMessageBroker.listen('watch-sections', (payload) {
      if (payload['name'] != null) {
        callback(TivioDataSection(payload['id'], payload['name'], this));
      }
    });

    return callFunction('watchSections', [channelId], (payload) => null);
  }

  unsubscribeToSections() {
    _callbackMessageBroker.removeListener('watch-sections');
  }

  subscribeToVideos(String sectionId, Function(TivioDataVideo) callback) {
    _callbackMessageBroker.listen('watch-videos', (payload) {
      if (payload['name'] != null) {
        callback(TivioDataVideo.fromDynamic(payload));
      }
    });

    return callFunction('watchVideos', [sectionId], (payload) => null);
  }

  unsubscribeToVideos() {
    _callbackMessageBroker.removeListener('watch-videos');
  }

  /// API 1.0.0

  /// Widget
  Future<TivioDataStream<TivioDataWidget>> subscribeToWidget({@required String widgetId, @required int count}) async {
    var _tivioWidgetDataStreamController = TivioDataStreamController<TivioDataWidget>();

    _callbackMessageBroker.listen('subscribe-to-widget:$widgetId', (payload) {
      _tivioWidgetDataStreamController.addSafely(TivioDataWidgetParser.parseWidget(payload));
    });

    await callFunction('subscribeToWidget', [widgetId], (payload) => null);

    return _tivioWidgetDataStreamController.tivioDataStream;
  }

  /// Channel

  Future<TivioDataStream<TivioDataPagination<TivioDataChannel>>> subscribeToChannelsInWidget({String widgetId, int count = 10}) async {
    _tivioChannelDataStreamController.reopenIfIsClosed();

    _callbackMessageBroker.listen('subscribe-to-channels-in-widget:$widgetId', (payload) {
      _tivioChannelDataStreamController.addSafely(TivioDataChannelParser.parsePaginatedChannels(payload));
    });

    await callFunction('subscribeToChannelsInWidget', [widgetId], (payload) => null);

    return _tivioChannelDataStreamController.tivioDataStream;
  }

  fetchNextChannelsInWidget({String widgetId, int count = 10}) async {
    _tivioChannelDataStreamController.addSafely(TivioDataPagination([], false));
  }

  Future<TivioDataStream<TivioDataChannel>> subscribeToChannel({@required String channelId}) async {
    var _tivioChannelDataStreamController = TivioDataStreamController<TivioDataChannel>();

    _callbackMessageBroker.listen('subscribe-to-channel:$channelId', (payload) {
      _tivioChannelDataStreamController.addSafely(TivioDataChannelParser.parseChannel(payload));
    });

    await callFunction('subscribeToChannel', [channelId], (payload) => null);

    return _tivioChannelDataStreamController.tivioDataStream;
  }


  /// Section

  Future<TivioDataStream<TivioDataPagination<TivioDataSection>>> subscribeToSectionsInChannel({@required String channelId, int count = 10}) async {
    _tivioSectionDataStreamController.reopenIfIsClosed();

    _callbackMessageBroker.listen('subscribe-to-sections-in-channel:$channelId', (payload) {
      _tivioSectionDataStreamController.addSafely(TivioDataSectionParser.parsePaginatedSections(payload));
    });

    await callFunction('subscribeToSectionsInChannel', [channelId], (payload) => null);

    return _tivioSectionDataStreamController.tivioDataStream;
  }

  fetchNextSectionsInChannel() {
    _tivioChannelDataStreamController.addSafely(TivioDataPagination([], false));
  }

  Future<TivioDataStream<TivioDataSection>> subscribeToSection({@required String sectionId}) async {
    var _tivioChannelDataStreamController = TivioDataStreamController<TivioDataSection>();

    _callbackMessageBroker.listen('subscribe-to-section:$sectionId', (payload) {
      _tivioChannelDataStreamController.addSafely(TivioDataSectionParser.parseSection(payload));
    });

    await callFunction('subscribeToSection', [sectionId], (payload) => null);

    return _tivioChannelDataStreamController.tivioDataStream;
  }


  /// Videos

  Future<TivioDataStream<TivioDataPagination<TivioDataVideo>>> subscribeToVideosInSection({@required String sectionId, @required int count}) async {
    _tivioVideoDataStreamController.reopenIfIsClosed();

    _callbackMessageBroker.listen('subscribe-to-videos-in-section:$sectionId', (payload) {
      _tivioVideoDataStreamController.addSafely(TivioDataVideoParser.parsePaginatedVideos(payload));
    });

    await callFunction('subscribeToVideosInSection', [sectionId], (payload) => null);

    return _tivioVideoDataStreamController.tivioDataStream;
  }

  fetchNextVideosInSection() {
    _tivioVideoDataStreamController.addSafely(TivioDataPagination([], false));
  }

  Future<TivioDataStream<TivioDataVideo>> subscribeToVideo({@required String videoId}) async {
    var _tivioVideoDataStreamController = TivioDataStreamController<TivioDataVideo>();

    _callbackMessageBroker.listen('subscribe-to-video:$videoId', (payload) {
      _tivioVideoDataStreamController.addSafely(TivioDataVideoParser.parseVideo(payload));
    });

    await callFunction('subscribeToVideo', [videoId], (payload) => null);

    return _tivioVideoDataStreamController.tivioDataStream;
  }
}
