import 'bridge.dart';

class TivioDataWidget {
  String id;
  String name;
  Bridge _bridge;
  List<TivioDataChannel> channels = [];

  TivioDataWidget(String id, String name, Bridge bridge, {List<TivioDataChannel> channels}) {
    this.id = id;
    this.name = name;
    _bridge = bridge;

    if (channels != null) {
      this.channels = channels;
    }
  }

  Future<List<TivioDataVideo>> getRecentVideos() {
    return this._bridge.widgetGetRecentVideos(this.id);
  }

  subscribeToChannels(Function(TivioDataChannel) callback) {
    return this._bridge.subscribeToChannels(this.id, callback);
  }

  unsubscribeToChannels() {
    // TD add support of unsubscribing concrete callback
    this._bridge.unsubscribeToChannels();
  }

  getNextChannels({int count = 10}) {}

  static TivioDataWidget fromDynamic(Map data, Bridge bridge) {
    var channels = data['channels']
        .map<TivioDataChannel>(
          (channel) => TivioDataChannel.fromDynamic(
            channel,
            bridge,
          ),
        )
        .toList();

    return TivioDataWidget(data['id'], data['name'], bridge, channels: channels);
  }
}

class _ChannelImages {
  String logo;
  String cover;

  _ChannelImages(String logo, String cover) {
    this.logo = logo;
    this.cover = cover;
  }
}

class TivioDataChannel {
  String id;
  String name;
  _ChannelImages images;
  Bridge _bridge;
  List<TivioDataSection> sections = [];

  TivioDataChannel(String id, String name, String logo, String cover, Bridge bridge,
      {List<TivioDataSection> sections}) {
    this.id = id;
    this.name = name;
    this.images = new _ChannelImages(logo, cover);
    _bridge = bridge;

    if (sections != null) {
      this.sections = sections;
    }
  }

  Future<List<TivioDataVideo>> getRecentVideos() {
    return this._bridge.channelGetRecentVideos(this.id);
  }

  subscribeToSections(Function(TivioDataSection) callback) {
    return this._bridge.subscribeToSections(this.id, callback);
  }

  unsubscribeToSections() {
    this._bridge.unsubscribeToSections();
  }

  static TivioDataChannel fromDynamic(Map data, Bridge bridge) {
    var sections = data['sections']
        .map<TivioDataSection>(
          (section) => TivioDataSection.fromDynamic(
            section,
            bridge,
          ),
        )
        .toList();

    return TivioDataChannel(data['id'], data['name'], data['headerLogo'], data['cover'], bridge, sections: sections);
  }
}

class TivioDataSection {
  String id;
  String name;
  Bridge _bridge;
  List<TivioDataVideo> videos = [];

  TivioDataSection(String id, String name, Bridge bridge, {List<TivioDataVideo> videos}) {
    this.id = id;
    this.name = name;
    _bridge = bridge;

    if (videos != null) {
      this.videos = videos;
    }
  }

  subscribeToVideos(Function(TivioDataVideo) callback) {
    return _bridge.subscribeToVideos(this.id, callback);
  }

  unsubscribeToVideos() {
    _bridge.unsubscribeToVideos();
  }

  static TivioDataSection fromDynamic(Map data, Bridge bridge) {
    var videos = data['videos']
        .map<TivioDataVideo>(
          (video) => TivioDataVideo.fromDynamic(video),
        )
        .toList();

    return TivioDataSection(data['id'], data['name'], bridge, videos: videos);
  }
}

class _Url {
  String hls;

  _Url(this.hls);
}

class _VideoImages {
  String cover;

  _VideoImages(this.cover);
}

class _Monetization {
  num price;
  String type;

  _Monetization(this.price, this.type);
}

class TivioDataPlaylistItem {
  String name;
  _Url url;
  Duration duration;

  TivioDataPlaylistItem(String name, String url, Duration duration) {
    this.name = name;
    this.url = _Url(url);
    this.duration = duration;
  }
}

class TivioDataVideo {
  String id;
  String name;
  Duration duration;
  DateTime created;
  _Url url;
  _VideoImages images;
  bool isPlaylist;
  List<TivioDataPlaylistItem> playlistItems = [];
  _Monetization monetization;

  TivioDataVideo(
    String id,
    String name,
    Duration duration,
    DateTime created,
    String url,
    String cover,
    bool isPlaylist,
    List<TivioDataPlaylistItem> playlistItems,
    Map monetization,
  ) {
    this.id = id;
    this.name = name;
    this.duration = duration;
    this.created = created;
    this.url = _Url(url);
    this.images = new _VideoImages(cover);
    this.isPlaylist = isPlaylist;
    this.playlistItems = playlistItems;
    if (monetization != null) {
      this.monetization = _Monetization(monetization['price'], monetization['type']);
    }
  }

  static TivioDataVideo fromDynamic(Map data) {
    List<TivioDataPlaylistItem> playlist = [];

    for (var playlistItem in data['chapters']) {
      playlist.add(TivioDataPlaylistItem(
        playlistItem['name'],
        playlistItem['uri'],
        Duration(milliseconds: playlistItem['duration'] ?? 0),
      ));
    }

    return TivioDataVideo(
      data['id'],
      data['name'],
      Duration(milliseconds: data['duration'] ?? 0),
      (data['created'] != null) ? DateTime.parse(data['created']) : null,
      data['url'] ?? '',
      data['cover'] ?? '',
      data['isPlaylist'] ?? false,
      playlist,
      data['monetization'],
    );
  }

  static List<TivioDataVideo> fromDynamicList(List data) {
    List<TivioDataVideo> videos = [];
    data.forEach((record) {
      videos.add(TivioDataVideo.fromDynamic(record));
    });

    return videos;
  }
}

class TivioDataPagination<T> {
  List<T> records;
  bool hasNextPage;

  TivioDataPagination(List<T> records, bool hasNextPage) {
    this.records = records;
    this.hasNextPage = hasNextPage;
  }
}
